# Generated by Django 2.1.8 on 2019-04-23 09:32

import uuid

import django.contrib.postgres.fields
import django.db.models.deletion
from django.db import migrations, models

import vng_api_common.fields
import vng_api_common.models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Applicatie",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "uuid",
                    models.UUIDField(
                        default=uuid.uuid4,
                        help_text="Unique resource identifier (UUID4)",
                        unique=True,
                    ),
                ),
                (
                    "client_ids",
                    django.contrib.postgres.fields.ArrayField(
                        base_field=models.CharField(max_length=50),
                        help_text="Comma-separated list of identifiers used for authentication",
                        size=None,
                    ),
                ),
                (
                    "label",
                    models.CharField(
                        help_text="A human readable representation of the application",
                        max_length=100,
                    ),
                ),
                (
                    "heeft_alle_autorisaties",
                    models.BooleanField(
                        default=False, help_text="Globally allows everything"
                    ),
                ),
            ],
            bases=(vng_api_common.models.APIMixin, models.Model),
        ),
        migrations.CreateModel(
            name="Autorisatie",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "component",
                    models.CharField(
                        help_text="Name of the component to authorize", max_length=50
                    ),
                ),
                (
                    "zaaktype",
                    models.URLField(
                        help_text="Url of the zaaktype that is allowed", max_length=1000
                    ),
                ),
                (
                    "scopes",
                    django.contrib.postgres.fields.ArrayField(
                        base_field=models.CharField(max_length=100),
                        help_text="Comma-separated list of identifiers used for authentication",
                        size=None,
                    ),
                ),
                (
                    "max_vertrouwelijkheidaanduiding",
                    vng_api_common.fields.VertrouwelijkheidsAanduidingField(
                        choices=[
                            ("openbaar", "OPENBAAR"),
                            ("beperkt openbaar", "BEPERKT OPENBAAR"),
                            ("intern", "INTERN"),
                            ("zaakvertrouwelijk", "ZAAKVERTROUWELIJK"),
                            ("vertrouwelijk", "VERTROUWELIJK"),
                            ("confidentieel", "CONFIDENTIEEL"),
                            ("geheim", "GEHEIM"),
                            ("zeer geheim", "ZEER GEHEIM"),
                        ],
                        help_text="Maximum level of confidentiality that is allowed",
                        max_length=20,
                    ),
                ),
                (
                    "applicatie",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="autorisaties",
                        to="authorizations.Applicatie",
                    ),
                ),
            ],
            bases=(vng_api_common.models.APIMixin, models.Model),
        ),
    ]



import json
import logging
import re
from typing import Callable

from django.conf import settings
from django.core.exceptions import ValidationError
from django.core.validators import RegexValidator
from django.utils import timezone
from django.utils.deconstruct import deconstructible
from django.utils.module_loading import import_string
from django.utils.translation import ugettext_lazy as _

import requests
from rest_framework import serializers, validators

from .constants import RSIN_LENGTH
from .oas import fetcher, obj_has_shape

logger = logging.getLogger(__name__)


WORD_REGEX = re.compile(r"[\w\-]+$", re.ASCII)


@deconstructible
class AlphanumericExcludingDiacritic:
    """
    Alle alfanumerieke tekens m.u.v. diacrieten.

    RGBZ heeft hier een vreemde definitie voor. De oorsprong is dat dit gek is
    voor bestandsnamen, en dus speciale karakters uitgesloten worden.
    """

    def __init__(self, start=0):
        self.start = start

    def __call__(self, value):
        stripped_value = value[self.start :]
        match = WORD_REGEX.match(stripped_value)
        if not match:
            raise ValidationError(
                'Waarde "{0}" mag geen diakrieten of non-ascii tekens bevatten{1}'.format(
                    value,
                    " na de eerste {0} karakters".format(self.start)
                    if self.start
                    else "",
                )
            )

    def __eq__(self, other):
        return (
            isinstance(other, AlphanumericExcludingDiacritic)
            and self.start == other.start
        )


# Default validator for entire string.
alphanumeric_excluding_diacritic = AlphanumericExcludingDiacritic()


def validate_non_negative_string(value):
    """
    Validate a string containing a integer to be non-negative.
    """
    error = False
    try:
        n = int(value)
    except ValueError:
        error = True
    if error or n < 0:
        raise ValidationError("De waarde moet een niet-negatief getal zijn.")


validate_digits = RegexValidator(
    regex="^[0-9]+$", message="Waarde moet numeriek zijn.", code="only-digits"
)


def validate_rsin(value):
    """
    Validates that a string value is a valid RSIN number by applying the
    '11-proef' checking.

    :param value: String object representing a presumably good RSIN number.
    """
    # Initial sanity checks.
    validate_digits(value)
    if len(value) != RSIN_LENGTH:
        raise ValidationError(
            "RSIN moet %s tekens lang zijn." % RSIN_LENGTH, code="invalid-length"
        )

    # 11-proef check.
    total = 0
    for multiplier, char in enumerate(reversed(value), start=1):
        if multiplier == 1:
            total += -multiplier * int(char)
        else:
            total += multiplier * int(char)

    if total % 11 != 0:
        raise ValidationError("Onjuist RSIN nummer.", code="invalid")


class URLValidator:
    """
    Validate that the URL actually resolves to a HTTP 200

    Any extra init kwargs are passed down to the underlying ``link_fetcher``

    :param get_auth: a callable returning appropriate headers to authenticate
      against the remote.
    """

    message = _(
        "The URL {url} responded with HTTP {status_code}. Please provide a valid URL."
    )
    code = "bad-url"

    def __init__(self, get_auth: Callable = None, **extra):
        self.get_auth = get_auth
        self.extra = extra

    def __call__(self, value: str):
        link_fetcher = import_string(settings.LINK_FETCHER)

        extra = self.extra.copy()

        # Handle auth for the remote URL
        if self.get_auth:
            auth_headers = self.get_auth(value)
            if "headers" not in self.extra:
                extra["headers"] = {}
            extra["headers"].update(auth_headers)

        try:
            response = link_fetcher(value, **extra)
        except Exception as exc:
            raise serializers.ValidationError(
                _("The URL {url} could not be fetched. Exception: {exc}").format(
                    url=value, exc=exc
                ),
                code=self.code,
            )

        if response.status_code != 200:
            raise serializers.ValidationError(
                self.message.format(status_code=response.status_code, url=value),
                code=self.code,
            )

        # return the response for post-processing
        return response


class ResourceValidator(URLValidator):
    """
    Validate that the URL resolves to an instance of the external resource.

    :param resource: name of the resource, e.g. 'zaak'
    :param oas_schema: URL to the schema to validate the response object shape
      against. Must be a YAML OAS 3.0.x spec.
    """

    # Name mangling is applied to these attributes to avoid formatting issues
    # that occur when overriding the superclass attributes
    __message = _(
        "The URL {url} resource did not look like a(n) `{resource}`. Please provide a valid URL."
    )
    __code = "invalid-resource"

    def __init__(self, resource: str, oas_schema: str, *args, **kwargs):
        self.resource = resource
        self.oas_schema = oas_schema
        super().__init__(*args, **kwargs)

    def __call__(self, url: str):
        response = super().__call__(url)

        # at this point, we know the URL actually exists
        try:
            obj = response.json()
        except json.JSONDecodeError as exc:
            logger.info(
                "URL %s doesn't seem to point to a JSON endpoint", url, exc_info=1
            )
            raise serializers.ValidationError(
                self.__message.format(url=url, resource=self.resource), code=self.__code
            )

        # check if the shape matches
        schema = fetcher.fetch(self.oas_schema)
        if not obj_has_shape(obj, schema, self.resource):
            logger.info(
                "URL %s doesn't seem to point to a valid shape", url, exc_info=1
            )
            raise serializers.ValidationError(
                self.__message.format(url=url, resource=self.resource), code=self.__code
            )

        return obj


class InformatieObjectUniqueValidator(validators.UniqueTogetherValidator):
    def __init__(self, parent_field, field: str):
        self.parent_field = parent_field
        self.field = field
        super().__init__(None, (parent_field, field))

    def set_context(self, serializer_field):
        serializer = serializer_field.parent
        super().set_context(serializer)

        self.queryset = serializer.Meta.model._default_manager.all()
        self.parent_object = serializer.context["parent_object"]

    def __call__(self, informatieobject: str):
        attrs = {self.parent_field: self.parent_object, self.field: informatieobject}
        super().__call__(attrs)


class ObjectInformatieObjectValidator:
    """
    Validate that the INFORMATIEOBJECT is linked already in the DRC.
    """

    message = _(
        "Het informatieobject is in het DRC nog niet gerelateerd aan dit object."
    )
    code = "inconsistent-relation"

    def set_context(self, serializer):
        """
        This hook is called by the serializer instance,
        prior to the validation call being made.
        """
        self.parent_object = serializer.context["parent_object"]
        self.request = serializer.context["request"]

    def __call__(self, informatieobject: str):
        from .models import APICredential

        object_url = self.parent_object.get_absolute_api_url(self.request)

        # dynamic so that it can be mocked in tests easily
        Client = import_string(settings.ZDS_CLIENT_CLASS)
        client = Client.from_url(informatieobject)
        client.auth = APICredential.get_auth(informatieobject)
        try:
            oios = client.list(
                "objectinformatieobject",
                query_params={
                    "informatieobject": informatieobject,
                    "object": object_url,
                },
            )
        except requests.HTTPError as exc:
            raise serializers.ValidationError(
                exc.args[0], code="relation-validation-error"
            ) from exc

        if len(oios) == 0:
            raise serializers.ValidationError(self.message, code=self.code)


@deconstructible
class UntilNowValidator:
    """
    Validate a datetime to not be in the future.

    This means that `now` is included.
    """

    message = _("Ensure this value is not in the future.")
    code = "future_not_allowed"

    @property
    def limit_value(self):
        return timezone.now()

    def __call__(self, value):
        if value > self.limit_value:
            raise ValidationError(self.message, code=self.code)

    def __eq__(self, other):
        return (
            isinstance(other, self.__class__)
            and self.message == other.message
            and self.code == other.code
        )


class UntilTodayValidator(UntilNowValidator):
    @property
    def limit_value(self):
        limit_value = super().limit_value
        return limit_value.date()


class UniekeIdentificatieValidator:
    """
    Valideer dat de identificatie binnen de organisatie uniek is.

    Indien de identificatie niet expliciet opgegeven is, wordt ervan uitgegaan
    dat de identificatie-generator uniciteit garandeert.

    :param organisatie_field: naam van het veld dat de organisatie RSIN bevat
    :param identificatie_field: naam van het veld dat de identificatie bevat
    """

    message = _("Deze identificatie bestaat al binnen de organisatie")
    code = "identificatie-niet-uniek"

    def __init__(self, organisatie_field: str, identificatie_field="identificatie"):
        self.organisatie_field = organisatie_field
        self.identificatie_field = identificatie_field

    def set_context(self, serializer):
        """
        This hook is called by the serializer instance,
        prior to the validation call being made.
        """
        # Determine the existing instance, if this is an update operation.
        self.instance = getattr(serializer, "instance", None)
        self.model = serializer.Meta.model

    def __call__(self, attrs: dict):
        identificatie = attrs.get(self.identificatie_field)
        if not identificatie:
            # identification is being generated, and the generation checks for
            # uniqueness
            return

        organisatie = attrs.get(self.organisatie_field)
        pk = self.instance.pk if self.instance else None

        # if we're updating an instance, setting the current values will not
        # trigger an error because the instance-to-be-updated is excluded from
        # the queryset. If either bronorganisatie or identificatie changes,
        # and it already exists, it will raise a validation error
        combination_exists = (
            self.model.objects
            # in case of an update, exclude the current object. for a create, this
            # will be None
            .exclude(pk=pk)
            .filter(
                **{
                    self.organisatie_field: organisatie,
                    self.identificatie_field: identificatie,
                }
            )
            .exists()
        )

        if combination_exists:
            raise serializers.ValidationError(
                {self.identificatie_field: self.message}, code=self.code
            )


class IsImmutableValidator:
    """
    Valideer dat de waarde van het veld niet wijzigt bij een update actie.
    """

    message = _("Dit veld mag niet gewijzigd worden.")
    code = "wijzigen-niet-toegelaten"

    def set_context(self, serializer_field):
        """
        This hook is called by the serializer instance,
        prior to the validation call being made.
        """
        # Determine the existing instance, if this is an update operation.
        self.serializer_field = serializer_field
        self.instance = getattr(serializer_field.parent, "instance", None)

    def __call__(self, new_value):
        # no instance -> it's not an update
        if not self.instance:
            return

        current_value = getattr(self.instance, self.serializer_field.field_name)

        if new_value != current_value:
            raise serializers.ValidationError(self.message, code=self.code)


class PublishValidator(ResourceValidator):
    """
    Validate that the URL actually resolves to a published resource (concept=False)
    """

    publish_message = _("The resource {url} is not published.")
    publish_code = "not-published"

    def __call__(self, url: str):
        response = super().__call__(url)

        if response.get("concept"):
            raise serializers.ValidationError(
                self.publish_message.format(url=url), code=self.publish_code
            )



from typing import Callable

from django.db import migrations

from ..constants import (
    ComponentTypes,
    RolOmschrijving,
    RolTypes,
    VertrouwelijkheidsAanduiding,
    ZaakobjectTypes,
)

VERTROUWELIJKHEIDSAANDUIDING_MAPPING = {
    "beperkt openbaar": VertrouwelijkheidsAanduiding.beperkt_openbaar,
    "zeer geheimp": VertrouwelijkheidsAanduiding.zeer_geheim,
}


ROLOMSCHRIJVING_MAPPING = {
    "Adviseur": RolOmschrijving.adviseur,
    "Behandelaar": RolOmschrijving.behandelaar,
    "Belanghebbende": RolOmschrijving.belanghebbende,
    "Beslisser": RolOmschrijving.beslisser,
    "Initiator": RolOmschrijving.initiator,
    "Klantcontacter": RolOmschrijving.klantcontacter,
    "Zaakcoördinator": RolOmschrijving.zaakcoordinator,
    "Mede-initiator": RolOmschrijving.medeinitiator,
}


ROLTYPES_MAPPING = {
    "Natuurlijk persoon": RolTypes.natuurlijk_persoon,
    "Niet-natuurlijk persoon": RolTypes.niet_natuurlijk_persoon,
    "Vestiging": RolTypes.vestiging,
    "Organisatorische eenheid": RolTypes.organisatorische_eenheid,
    "Medewerker": RolTypes.medewerker,
}


ZAAKOBJECTTYPES_MAPPING = {
    "enkelvoudigDocument": ZaakobjectTypes.enkelvoudig_document,
    "gemeentelijkeOpenbareRuimte": ZaakobjectTypes.gemeentelijke_openbare_ruimte,
    "kadastraleOnroerendeZaak": ZaakobjectTypes.kadastrale_onroerende_zaak,
    "maatschappelijkeActiviteit": ZaakobjectTypes.maatschappelijke_activiteit,
    "natuurlijkPersoon": ZaakobjectTypes.natuurlijk_persoon,
    "nietNatuurlijkPersoon": ZaakobjectTypes.niet_natuurlijk_persoon,
    "openbareRuimte": ZaakobjectTypes.openbare_ruimte,
    "organisatorischeEenheid": ZaakobjectTypes.organisatorische_eenheid,
    "terreinGebouwdObject": ZaakobjectTypes.terrein_gebouwd_object,
    "wozDeelobject": ZaakobjectTypes.woz_deelobject,
    "wozObject": ZaakobjectTypes.woz_object,
    "wozWaarde": ZaakobjectTypes.woz_waarde,
    "zakelijkRecht": ZaakobjectTypes.zakelijk_recht,
}


COMPONENTTYPES_MAPPING = {
    "AC": ComponentTypes.ac,
    "NRC": ComponentTypes.nrc,
    "ZRC": ComponentTypes.zrc,
    "ZTC": ComponentTypes.ztc,
    "DRC": ComponentTypes.drc,
    "BRC": ComponentTypes.brc,
}


def update_factory(model: str, field: str, mapping: dict) -> Callable:
    def operation(apps, schema_editor):
        Model = apps.get_model(model)
        for old_value, new_value in mapping.items():
            qs = Model.objects.filter(**{field: old_value})
            qs.update(**{field: new_value})

    return operation


class UpdateChoiceValues(migrations.RunPython):
    mapping: dict

    def __init__(self, model: str, field: str, *args, **kwargs):
        forward = update_factory(model, field, self.mapping)
        backwards = update_factory(
            model, field, {value: key for key, value in self.mapping.items()}
        )

        super().__init__(forward, backwards, *args, **kwargs)


class VertrouwelijkheidsaanduidingUpdate(UpdateChoiceValues):
    mapping = VERTROUWELIJKHEIDSAANDUIDING_MAPPING


class RolomschrijvingUpdate(UpdateChoiceValues):
    mapping = ROLOMSCHRIJVING_MAPPING


class RoltypesUpdate(UpdateChoiceValues):
    mapping = ROLTYPES_MAPPING


class ZaakobjecttypesUpdate(UpdateChoiceValues):
    mapping = ZAAKOBJECTTYPES_MAPPING


class ComponenttypesUpdate(UpdateChoiceValues):
    mapping = COMPONENTTYPES_MAPPING



from rest_framework.exceptions import NotAcceptable
from rest_framework.renderers import BrowsableAPIRenderer

from .exceptions import PreconditionFailed

HEADER_ACCEPT = "Accept-Crs"
HEADER_CONTENT = "Content-Crs"

DEFAULT_CRS = "EPSG:4326"


def extract_header(request, name: str) -> str:
    _header = name.replace("-", "_").upper()
    header = f"HTTP_{_header}"
    return request.META.get(header)


class GeoMixin:
    """
    GeoJSON viewset mixin.
    """

    @property
    def default_response_headers(self):
        headers = super().default_response_headers
        headers["Content-Crs"] = DEFAULT_CRS
        return headers

    def initial(self, request, *args, **kwargs):
        super().initial(request, *args, **kwargs)
        self.perform_crs_negotation(request)

    def perform_crs_negotation(self, request):
        # don't cripple the browsable API...
        if isinstance(request.accepted_renderer, BrowsableAPIRenderer):
            return

        # methods with request bodies need to have the CRS specified
        if request.method.lower() in ("post", "put", "patch"):
            content_crs = extract_header(request, HEADER_CONTENT)
            if content_crs is None:
                raise PreconditionFailed(detail=f"'{HEADER_CONTENT}' header ontbreekt")
            if content_crs != DEFAULT_CRS:
                raise NotAcceptable(detail=f"CRS '{content_crs}' is niet ondersteund")

        if request.method.lower() == "delete":
            return

        # client must indicate which CRS they want in the response
        requested_crs = extract_header(request, HEADER_ACCEPT)
        if requested_crs is None:
            raise PreconditionFailed(detail=f"'{HEADER_ACCEPT}' header ontbreekt")

        if requested_crs != DEFAULT_CRS:
            raise NotAcceptable(detail=f"CRS '{requested_crs}' is niet ondersteund")



import logging

from drf_yasg import openapi
from drf_yasg.inspectors.base import NotHandled
from drf_yasg.inspectors.field import FieldInspector, InlineSerializerInspector
from rest_framework import serializers

from ..serializers import GegevensGroepSerializer

logger = logging.getLogger(__name__)


TYPES_MAP = {
    str: openapi.TYPE_STRING,
    int: openapi.TYPE_INTEGER,
    bool: openapi.TYPE_BOOLEAN,
}


class ReadOnlyFieldInspector(FieldInspector):
    """
    Provides conversion for derived ReadOnlyField from model fields.

    This inspector looks at the type hint to determine the type/format of
    a model property.
    """

    def field_to_swagger_object(
        self, field, swagger_object_type, use_references, **kwargs
    ):
        SwaggerType, ChildSwaggerType = self._get_partial_types(
            field, swagger_object_type, use_references, **kwargs
        )

        if (
            isinstance(field, serializers.ReadOnlyField)
            and swagger_object_type == openapi.Schema
        ):
            prop = getattr(field.parent.Meta.model, field.source)
            if not isinstance(prop, property):
                return NotHandled

            return_type = prop.fget.__annotations__.get("return")
            if return_type is None:  # no type annotation, too bad...
                logger.debug(
                    "Missing return type annotation for prop %s on model %s",
                    field.source,
                    field.parent.Meta.model,
                )
                return NotHandled

            type_ = TYPES_MAP.get(return_type)
            if type_ is None:
                logger.debug("Missing type mapping for %r", return_type)

            return SwaggerType(type=type_ or openapi.TYPE_STRING)

        return NotHandled


class JSONFieldInspector(FieldInspector):
    def field_to_swagger_object(
        self, field, swagger_object_type, use_references, **kwargs
    ):  # pragma: no cover
        SwaggerType, ChildSwaggerType = self._get_partial_types(
            field, swagger_object_type, use_references, **kwargs
        )

        if (
            isinstance(field, serializers.JSONField)
            and swagger_object_type == openapi.Schema
        ):
            return SwaggerType(type=openapi.TYPE_OBJECT)

        return NotHandled


class HyperlinkedIdentityFieldInspector(FieldInspector):
    def field_to_swagger_object(
        self, field, swagger_object_type, use_references, **kwargs
    ):
        SwaggerType, ChildSwaggerType = self._get_partial_types(
            field, swagger_object_type, use_references, **kwargs
        )

        if (
            isinstance(field, serializers.HyperlinkedIdentityField)
            and swagger_object_type == openapi.Schema
        ):
            return SwaggerType(
                type=openapi.TYPE_STRING,
                format=openapi.FORMAT_URI,
                min_length=1,
                max_length=1000,
                description="URL-referentie naar dit object. Dit is de unieke identificatie en locatie van dit object.",
            )

        return NotHandled


class GegevensGroepInspector(FieldInspector):
    def process_result(self, result, method_name, obj, **kwargs):
        if not isinstance(result, openapi.Schema.OR_REF):
            return result

        if not isinstance(obj, GegevensGroepSerializer):
            return result

        if method_name != "field_to_swagger_object":
            return result

        if not obj.allow_null:
            return result

        schema = openapi.resolve_ref(result, self.components)
        schema.x_nullable = True

        return result



import logging

from django.db import transaction
from django.http import Http404

from rest_framework import viewsets

from ..compat import get_header
from ..constants import CommonResourceAction
from ..permissions import AuthScopesRequired
from ..utils import get_uuid_from_path
from ..viewsets import NestedViewSetMixin
from .api.scopes import SCOPE_AUDITTRAILS_LEZEN
from .api.serializers import AuditTrailSerializer
from .models import AuditTrail

logger = logging.getLogger(__name__)


class AuditTrailMixin:
    audit = None

    def get_audittrail_main_object_url(self, data, main_resource):
        """
        Retrieve the URL that points to the main resource
        """
        if hasattr(self, "audittrail_main_resource_key"):
            return data[self.audittrail_main_resource_key]
        return data[main_resource]

    def create_audittrail(
        self,
        status_code,
        action,
        version_before_edit,
        version_after_edit,
        unique_representation,
    ):
        """
        Create the audittrail for the action that has been carried out.
        """
        data = version_after_edit if version_after_edit else version_before_edit
        if self.basename == self.audit.main_resource:
            main_object = data["url"]
        else:
            main_object = self.get_audittrail_main_object_url(
                data, self.audit.main_resource
            )

        applications = self.request.jwt_auth.applicaties
        if len(applications) > 1:
            logger.warning(
                "Unexpectedly found %d applications, expected at most one",
                len(applications),
            )

        if applications:
            application = applications[0]
            app_id, app_presentation = str(application.uuid), application.label
        else:
            app_id = get_header(self.request, "X-NLX-Request-Application-Id")
            app_presentation = app_id  # we don't have any extra information...

        user_id = self.request.jwt_auth.payload.get("user_id", "")
        if not user_id:
            user_id = get_header(self.request, "X-NLX-Request-User-Id") or ""

        request_id = get_header(self.request, "X-NLX-Request-Id") or ""

        toelichting = get_header(self.request, "X-Audit-Toelichting") or ""

        trail = AuditTrail(
            bron=self.audit.component_name,
            request_id=request_id,
            applicatie_id=app_id,
            applicatie_weergave=app_presentation,
            actie=action,
            actie_weergave=CommonResourceAction.labels.get(action, ""),
            gebruikers_id=user_id,
            gebruikers_weergave=self.request.jwt_auth.payload.get(
                "user_representation", ""
            ),
            resultaat=status_code,
            hoofd_object=main_object,
            resource=self.basename,
            resource_url=data["url"],
            toelichting=toelichting,
            resource_weergave=unique_representation,
            oud=version_before_edit,
            nieuw=version_after_edit,
        )
        trail.save()


class AuditTrailCreateMixin(AuditTrailMixin):
    def get_audittrail_instance(self, response):
        zaak_uuid = get_uuid_from_path(response.data["url"])
        instance = self.get_queryset().get(uuid=zaak_uuid)
        return instance

    def create(self, request, *args, **kwargs):
        response = super().create(request, *args, **kwargs)
        instance = self.get_audittrail_instance(response)
        self.create_audittrail(
            response.status_code,
            CommonResourceAction.create,
            version_before_edit=None,
            version_after_edit=response.data,
            unique_representation=instance.unique_representation(),
        )
        return response


class AuditTrailUpdateMixin(AuditTrailMixin):
    def update(self, request, *args, **kwargs):
        # Retrieve the data stored in the object before updating
        instance = self.get_object()
        serializer = self.get_serializer(instance)
        version_before_edit = serializer.data

        action = (
            CommonResourceAction.partial_update
            if kwargs.get("partial", False)
            else CommonResourceAction.update
        )

        response = super().update(request, *args, **kwargs)
        self.create_audittrail(
            response.status_code,
            action,
            version_before_edit=version_before_edit,
            version_after_edit=response.data,
            unique_representation=instance.unique_representation(),
        )
        return response


class AuditTrailDestroyMixin(AuditTrailMixin):
    def destroy(self, request, *args, **kwargs):
        # Retrieve the data stored in the object before updating
        instance = self.get_object()
        serializer = self.get_serializer(instance)
        version_before_edit = serializer.data

        # If the resource being deleted is the main resource, delete all the
        # audittrails associated with it
        if self.basename == self.audit.main_resource:
            with transaction.atomic():
                response = super().destroy(request, *args, **kwargs)
                self._destroy_related_audittrails(version_before_edit["url"])
                return response
        else:
            response = super().destroy(request, *args, **kwargs)
            self.create_audittrail(
                response.status_code,
                CommonResourceAction.destroy,
                version_before_edit=version_before_edit,
                version_after_edit=None,
                unique_representation=instance.unique_representation(),
            )
            return response

    def _destroy_related_audittrails(self, main_object_url):
        AuditTrail.objects.filter(hoofd_object=main_object_url).delete()


class AuditTrailViewsetMixin(
    AuditTrailCreateMixin, AuditTrailUpdateMixin, AuditTrailDestroyMixin
):
    pass


class AuditTrailViewSet(NestedViewSetMixin, viewsets.ReadOnlyModelViewSet):
    """
    ViewSet that shows the Audit trails for a resource (e.g. a Zaak)

    In order to create an AuditTrailViewSet for a specific resource, this class
    must be inherited from in the viewsets of the component where this resource
    lives, and the `main_resource_lookup_field` must be set to the identifier
    of the resource (usually uuid)

    Example usage for a AuditTrailViewSet for the `Zaak`-resource:

        class ZaakAuditTrailViewset(AuditTrailViewset):
            main_resource_lookup_field = 'zaak_uuid'
    """

    queryset = AuditTrail.objects.all().order_by("aanmaakdatum")
    serializer_class = AuditTrailSerializer
    lookup_field = "uuid"
    permission_classes = (AuthScopesRequired,)
    required_scopes = {
        "list": SCOPE_AUDITTRAILS_LEZEN,
        "retrieve": SCOPE_AUDITTRAILS_LEZEN,
    }

    main_resource_lookup_field = None  # Must be overwritten by subclasses

    def get_queryset(self):
        qs = super().get_queryset()
        identifier = self.kwargs.get(self.main_resource_lookup_field)
        if identifier:
            filtered = qs.filter(hoofd_object__contains=identifier)
            if filtered.exists():
                return filtered
            else:
                raise Http404
        return qs



from collections import OrderedDict

from drf_yasg import openapi
from drf_yasg.inspectors import FieldInspector, NotHandled
from rest_framework import serializers
from rest_framework_gis.fields import GeometryField

from ..geo import DEFAULT_CRS, HEADER_ACCEPT, HEADER_CONTENT

REF_NAME_GEOJSON_GEOMETRY = "GeoJSONGeometry"


def register_geojson(definitions):
    Geometry = openapi.Schema(
        type=openapi.TYPE_OBJECT,
        title="Geometry",
        description="GeoJSON geometry",
        required=["type"],
        externalDocs=OrderedDict(url="https://tools.ietf.org/html/rfc7946#section-3.1"),
        properties=OrderedDict(
            (
                (
                    "type",
                    openapi.Schema(
                        type=openapi.TYPE_STRING,
                        enum=[
                            "Point",
                            "MultiPoint",
                            "LineString",
                            "MultiLineString",
                            "Polygon",
                            "MultiPolygon",
                            "Feature",
                            "FeatureCollection",
                            "GeometryCollection",
                        ],
                        description="The geometry type",
                    ),
                ),
            )
        ),
    )
    definitions.set("Geometry", Geometry)

    Point2D = openapi.Schema(
        type=openapi.TYPE_ARRAY,
        title="Point2D",
        description="A 2D point",
        items=openapi.Schema(type=openapi.TYPE_NUMBER),
        maxItems=2,
        minItems=2,
    )
    definitions.set("Point2D", Point2D)

    Point = openapi.Schema(
        type=openapi.TYPE_OBJECT,
        description="GeoJSON point geometry",
        externalDocs=OrderedDict(
            url="https://tools.ietf.org/html/rfc7946#section-3.1.2"
        ),
        allOf=[
            openapi.SchemaRef(definitions, "Geometry"),
            openapi.Schema(
                type=openapi.TYPE_OBJECT,
                required=["coordinates"],
                properties=OrderedDict(
                    (("coordinates", openapi.SchemaRef(definitions, "Point2D")),)
                ),
            ),
        ],
    )
    definitions.set("Point", Point)

    MultiPoint = openapi.Schema(
        type=openapi.TYPE_OBJECT,
        description="GeoJSON multi-point geometry",
        externalDocs=OrderedDict(
            url="https://tools.ietf.org/html/rfc7946#section-3.1.3"
        ),
        allOf=[
            openapi.SchemaRef(definitions, "Geometry"),
            openapi.Schema(
                type=openapi.TYPE_OBJECT,
                required=["coordinates"],
                properties=OrderedDict(
                    (
                        (
                            "coordinates",
                            openapi.Schema(
                                type=openapi.TYPE_ARRAY,
                                items=openapi.SchemaRef(definitions, "Point2D"),
                            ),
                        ),
                    )
                ),
            ),
        ],
    )
    definitions.set("MultiPoint", MultiPoint)

    LineString = openapi.Schema(
        type=openapi.TYPE_OBJECT,
        description="GeoJSON line-string geometry",
        externalDocs=OrderedDict(
            url="https://tools.ietf.org/html/rfc7946#section-3.1.4"
        ),
        allOf=[
            openapi.SchemaRef(definitions, "Geometry"),
            openapi.Schema(
                type=openapi.TYPE_OBJECT,
                required=["coordinates"],
                properties=OrderedDict(
                    (
                        (
                            "coordinates",
                            openapi.Schema(
                                type=openapi.TYPE_ARRAY,
                                items=openapi.SchemaRef(definitions, "Point2D"),
                                minItems=2,
                            ),
                        ),
                    )
                ),
            ),
        ],
    )
    definitions.set("LineString", LineString)

    MultiLineString = openapi.Schema(
        type=openapi.TYPE_OBJECT,
        description="GeoJSON multi-line-string geometry",
        externalDocs=OrderedDict(
            url="https://tools.ietf.org/html/rfc7946#section-3.1.5"
        ),
        allOf=[
            openapi.SchemaRef(definitions, "Geometry"),
            openapi.Schema(
                type=openapi.TYPE_OBJECT,
                required=["coordinates"],
                properties=OrderedDict(
                    (
                        (
                            "coordinates",
                            openapi.Schema(
                                type=openapi.TYPE_ARRAY,
                                items=openapi.Schema(
                                    type=openapi.TYPE_ARRAY,
                                    items=openapi.SchemaRef(definitions, "Point2D"),
                                ),
                            ),
                        ),
                    )
                ),
            ),
        ],
    )
    definitions.set("MultiLineString", MultiLineString)

    Polygon = openapi.Schema(
        type=openapi.TYPE_OBJECT,
        description="GeoJSON polygon geometry",
        externalDocs=OrderedDict(
            url="https://tools.ietf.org/html/rfc7946#section-3.1.6"
        ),
        allOf=[
            openapi.SchemaRef(definitions, "Geometry"),
            openapi.Schema(
                type=openapi.TYPE_OBJECT,
                required=["coordinates"],
                properties=OrderedDict(
                    (
                        (
                            "coordinates",
                            openapi.Schema(
                                type=openapi.TYPE_ARRAY,
                                items=openapi.Schema(
                                    type=openapi.TYPE_ARRAY,
                                    items=openapi.SchemaRef(definitions, "Point2D"),
                                ),
                            ),
                        ),
                    )
                ),
            ),
        ],
    )
    definitions.set("Polygon", Polygon)

    MultiPolygon = openapi.Schema(
        type=openapi.TYPE_OBJECT,
        description="GeoJSON multi-polygon geometry",
        externalDocs=OrderedDict(
            url="https://tools.ietf.org/html/rfc7946#section-3.1.7"
        ),
        allOf=[
            openapi.SchemaRef(definitions, "Geometry"),
            openapi.Schema(
                type=openapi.TYPE_OBJECT,
                required=["coordinates"],
                properties=OrderedDict(
                    (
                        (
                            "coordinates",
                            openapi.Schema(
                                type=openapi.TYPE_ARRAY,
                                items=openapi.Schema(
                                    type=openapi.TYPE_ARRAY,
                                    items=openapi.Schema(
                                        type=openapi.TYPE_ARRAY,
                                        items=openapi.SchemaRef(definitions, "Point2D"),
                                    ),
                                ),
                            ),
                        ),
                    )
                ),
            ),
        ],
    )
    definitions.set("MultiPolygon", MultiPolygon)

    GeometryCollection = openapi.Schema(
        type=openapi.TYPE_OBJECT,
        description="GeoJSON multi-polygon geometry",
        externalDocs=OrderedDict(
            url="https://tools.ietf.org/html/rfc7946#section-3.1.8"
        ),
        allOf=[
            openapi.SchemaRef(definitions, "Geometry"),
            openapi.Schema(
                type=openapi.TYPE_OBJECT,
                required=["geometries"],
                properties=OrderedDict(
                    (
                        (
                            "geometries",
                            openapi.Schema(
                                type=openapi.TYPE_ARRAY,
                                items=openapi.SchemaRef(definitions, "Geometry"),
                            ),
                        ),
                    )
                ),
            ),
        ],
    )
    definitions.set("GeometryCollection", GeometryCollection)

    GeoJSONGeometry = openapi.Schema(
        title=REF_NAME_GEOJSON_GEOMETRY,
        type=openapi.TYPE_OBJECT,
        oneOf=[
            openapi.SchemaRef(definitions, "Point"),
            openapi.SchemaRef(definitions, "MultiPoint"),
            openapi.SchemaRef(definitions, "LineString"),
            openapi.SchemaRef(definitions, "MultiLineString"),
            openapi.SchemaRef(definitions, "Polygon"),
            openapi.SchemaRef(definitions, "MultiPolygon"),
            openapi.SchemaRef(definitions, "GeometryCollection"),
        ],
        discriminator="type",
    )
    definitions.set(REF_NAME_GEOJSON_GEOMETRY, GeoJSONGeometry)


class GeometryFieldInspector(FieldInspector):
    def field_to_swagger_object(
        self, field, swagger_object_type, use_references, **kwargs
    ):
        if not isinstance(field, GeometryField):
            return NotHandled

        definitions = self.components.with_scope(openapi.SCHEMA_DEFINITIONS)

        if not definitions.has("Geometry"):
            register_geojson(definitions)

        return openapi.SchemaRef(definitions, REF_NAME_GEOJSON_GEOMETRY)

    def has_geo_fields(self, serializer) -> bool:
        """
        Check if any of the serializer fields are a GeometryField.

        If the serializer has nested serializers, a depth-first search is done
        to check if the nested serializers has `GeometryField`\ s.
        """
        for field in serializer.fields.values():
            if isinstance(field, serializers.Serializer):
                has_nested_geo_fields = self.probe_inspectors(
                    self.field_inspectors,
                    "has_geo_fields",
                    field,
                    {"field_inspectors": self.field_inspectors},
                )
                if has_nested_geo_fields:
                    return True

            elif isinstance(field, (serializers.ListSerializer, serializers.ListField)):
                field = field.child

            if isinstance(field, GeometryField):
                return True

        return False

    def get_request_header_parameters(self, serializer):
        if not self.has_geo_fields(serializer):
            return []

        if self.method == "DELETE":
            return []

        # see also http://lyzidiamond.com/posts/4326-vs-3857 for difference
        # between coordinate system and projected coordinate system
        return [
            openapi.Parameter(
                name=HEADER_ACCEPT,
                type=openapi.TYPE_STRING,
                in_=openapi.IN_HEADER,
                required=True,
                description="Het gewenste 'Coordinate Reference System' (CRS) van de "
                "geometrie in het antwoord (response body). Volgens de "
                "GeoJSON spec is WGS84 de default (EPSG:4326 is "
                "hetzelfde als WGS84).",
                enum=[DEFAULT_CRS],
            ),
            openapi.Parameter(
                name=HEADER_CONTENT,
                type=openapi.TYPE_STRING,
                in_=openapi.IN_HEADER,
                required=True,
                description="Het 'Coordinate Reference System' (CRS) van de "
                "geometrie in de vraag (request body). Volgens de "
                "GeoJSON spec is WGS84 de default (EPSG:4326 is "
                "hetzelfde als WGS84).",
                enum=[DEFAULT_CRS],
            ),
        ]

    def get_response_headers(self, serializer, status=None):
        if not self.has_geo_fields(serializer):
            return None

        if int(status) != 200:
            return None

        return OrderedDict(
            (
                (
                    HEADER_CONTENT,
                    openapi.Schema(
                        type=openapi.TYPE_STRING,
                        enum=[DEFAULT_CRS],
                        description="Het 'Coordinate Reference System' (CRS) van de "
                        "antwoorddata. Volgens de GeoJSON spec is WGS84 de "
                        "default (EPSG:4326 is hetzelfde als WGS84).",
                    ),
                ),
            )
        )



import logging
import os

from django.apps import apps
from django.contrib.auth.models import User
from django.core.exceptions import ImproperlyConfigured
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.module_loading import import_string

from drf_yasg import openapi
from drf_yasg.app_settings import swagger_settings
from drf_yasg.management.commands import generate_swagger
from rest_framework.settings import api_settings

from ...schema import OpenAPISchemaGenerator
from ...version import get_major_version


class Table:
    def __init__(self, resource: str):
        self.resource = resource
        self.rows = []


class Row:
    def __init__(
        self,
        label: str,
        description: str,
        type: str,
        required: bool,
        create: bool,
        read: bool,
        update: bool,
        delete: bool,
    ):
        self.label = label
        self.description = description
        self.type = type
        self.required = required
        self.create = create
        self.read = read
        self.update = update
        self.delete = delete


class Command(generate_swagger.Command):
    """
    Patches to the provided command to modify the schema for ZDS needs.
    """

    leave_locale_alone = True

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument("--to-markdown-table", action="store_true")

        parser.add_argument(
            "--info", dest="info", default=None, help="Path to schema info object"
        )

        parser.add_argument(
            "--urlconf",
            dest="urlconf",
            default=None,
            help="Urlconf for schema generator",
        )

    def get_mock_request(self, *args, **kwargs):
        request = super().get_mock_request(*args, **kwargs)
        request.version = api_settings.DEFAULT_VERSION
        return request

    def write_schema(self, schema, stream, format):
        del schema.host
        del schema.schemes
        super().write_schema(schema, stream, format)

    # need to overwrite the generator class...
    def handle(
        self,
        output_file,
        overwrite,
        format,
        api_url,
        mock,
        user,
        private,
        info=None,
        urlconf=None,
        *args,
        **options,
    ):
        # disable logs of WARNING and below
        logging.disable(logging.WARNING)

        if info:
            info = import_string(info)
        else:
            info = getattr(swagger_settings, "DEFAULT_INFO", None)
        if not isinstance(info, openapi.Info):
            raise ImproperlyConfigured(
                'settings.SWAGGER_SETTINGS["DEFAULT_INFO"] should be an '
                "import string pointing to an openapi.Info object"
            )

        if not format:
            if os.path.splitext(output_file)[1] in (".yml", ".yaml"):
                format = "yaml"
        format = format or "json"

        api_root = reverse("api-root", kwargs={"version": get_major_version()})
        api_url = (
            api_url
            or swagger_settings.DEFAULT_API_URL  # noqa
            or f"http://example.com{api_root}"  # noqa
        )

        user = User.objects.get(username=user) if user else None
        mock = mock or private or (user is not None)
        if mock and not api_url:
            raise ImproperlyConfigured(
                "--mock-request requires an API url; either provide "
                "the --url argument or set the DEFAULT_API_URL setting"
            )

        request = self.get_mock_request(api_url, format, user) if mock else None

        generator = OpenAPISchemaGenerator(info=info, url=api_url, urlconf=urlconf)
        schema = generator.get_schema(request=request, public=not private)

        if output_file == "-":
            self.write_schema(schema, self.stdout, format)
        else:
            with open(output_file, "w", encoding="utf8") as stream:
                if options["to_markdown_table"]:
                    self.to_markdown_table(schema, stream)
                else:
                    self.write_schema(schema, stream, format)

    def to_markdown_table(self, schema, stream):
        template = "vng_api_common/api_schema_to_markdown_table.md"
        tables = []

        whitelist = [model._meta.object_name for model in apps.get_models()]

        for resource, definition in schema.definitions.items():
            if resource not in whitelist:
                continue

            if not hasattr(definition, "properties"):
                continue

            table = Table(resource)
            for field, _schema in definition.properties.items():
                if isinstance(_schema, openapi.SchemaRef):
                    continue
                required = (
                    hasattr(definition, "required") and field in definition.required
                )

                readonly = getattr(_schema, "readOnly", False)
                table.rows.append(
                    Row(
                        label=field,
                        description=getattr(_schema, "description", ""),
                        type=_schema.type,
                        required=required,
                        create=not readonly,
                        read=True,
                        update=not readonly,
                        delete=not readonly,
                    )
                )
            tables.append(table)

        markdown = render_to_string(template, context={"tables": tables})
        stream.write(markdown)



import uuid

from django.contrib.postgres.fields import JSONField
from django.contrib.postgres.indexes import GinIndex
from django.core.serializers.json import DjangoJSONEncoder
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.utils.translation import ugettext_lazy as _

from ..constants import ComponentTypes
from ..descriptors import GegevensGroepType


class AuditTrail(models.Model):
    uuid = models.UUIDField(
        unique=True,
        default=uuid.uuid4,
        help_text=_("Unieke identificatie van de audit regel."),
    )
    request_id = models.CharField(
        max_length=255,
        blank=True,
        help_text=_(
            'Een globaal "request" ID om een verzoek door het '
            "netwerk heen te traceren."
        ),
    )
    bron = models.CharField(
        max_length=50,
        help_text=_("De naam van het component waar de wijziging in is gedaan."),
        choices=ComponentTypes.choices,
    )
    actie = models.CharField(max_length=50, help_text=_("De uitgevoerde handeling."))
    actie_weergave = models.CharField(
        max_length=200, blank=True, help_text=_("Vriendelijke naam van de actie.")
    )
    resultaat = models.IntegerField(
        help_text=_(
            "HTTP status code van de API response van de uitgevoerde handeling."
        ),
        validators=[MinValueValidator(100), MaxValueValidator(599)],
    )
    hoofd_object = models.URLField(
        max_length=1000, help_text=_("De URL naar het hoofdobject van een component.")
    )
    resource = models.CharField(
        max_length=50, help_text=_("Het type resource waarop de actie gebeurde.")
    )
    resource_url = models.URLField(
        max_length=1000, help_text=_("De URL naar het object.")
    )
    aanmaakdatum = models.DateTimeField(
        auto_now=True, help_text=_("De datum waarop de handeling is gedaan.")
    )
    resource_weergave = models.CharField(
        max_length=200, help_text=_("Vriendelijke identificatie van het object.")
    )
    applicatie_id = models.CharField(
        max_length=100,
        blank=True,
        help_text=_("Unieke identificatie van de applicatie, binnen de organisatie."),
    )
    applicatie_weergave = models.CharField(
        max_length=200, blank=True, help_text=_("Vriendelijke naam van de applicatie.")
    )
    oud = JSONField(
        null=True,
        encoder=DjangoJSONEncoder,
        help_text=_(
            "Volledige JSON body van het object zoals dat bestond voordat de actie heeft plaatsgevonden."
        ),
    )
    nieuw = JSONField(
        null=True,
        encoder=DjangoJSONEncoder,
        help_text=_("Volledige JSON body van het object na de actie."),
    )

    gebruikers_id = models.CharField(
        max_length=255,
        blank=True,
        help_text=_(
            "Unieke identificatie van de gebruiker die binnen de "
            "organisatie herleid kan worden naar een persoon."
        ),
    )
    gebruikers_weergave = models.CharField(
        max_length=255, blank=True, help_text=_("Vriendelijke naam van de gebruiker.")
    )
    toelichting = models.TextField(
        _("toelichting"),
        blank=True,
        help_text=_("Toelichting waarom de handeling is uitgevoerd."),
    )
    wijzigingen = GegevensGroepType(
        {"oud": oud, "nieuw": nieuw}, optional=["oud", "nieuw"], none_for_empty=True
    )

    class Meta:
        indexes = [
            GinIndex(
                fields=["hoofd_object"],
                name="audittrail_hoofdobject_trgm",
                opclasses=["gin_trgm_ops"],
            )
        ]



# Generated by Django 2.0.13 on 2019-03-25 13:13

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("notifications", "0003_auto_20190319_1048")]

    operations = [
        migrations.AlterField(
            model_name="notificationsconfig",
            name="api_root",
            field=models.URLField(
                default="https://ref.tst.vng.cloud/nc/api/v1",
                unique=True,
                verbose_name="api root",
            ),
        )
    ]



# Generated by Django 2.0.13 on 2019-05-20 12:38

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("audittrails", "0003_auto_20190517_0844")]

    operations = [
        migrations.AddField(
            model_name="audittrail",
            name="gebruikers_id",
            field=models.CharField(
                blank=True,
                help_text="Unieke identificatie van de gebruiker die binnen de organisatie herleid kan worden naar een persoon",
                max_length=255,
            ),
        ),
        migrations.AddField(
            model_name="audittrail",
            name="gebruikers_weergave",
            field=models.CharField(
                blank=True,
                help_text="Vriendelijke naam van de gebruiker",
                max_length=255,
            ),
        ),
    ]



"""
Optional app to manage notifications.

Components can send and receive notifications. This app provides hooks to
configure and register webhooks for notifications, and a generic APIView to
receive incoming notifications.
"""
default_app_config = "vng_api_common.notifications.apps.NotificationsConfig"



# Generated by Django 2.1.8 on 2019-05-06 12:12

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("authorizations", "0006_auto_20190506_0901")]

    operations = [
        migrations.AddField(
            model_name="autorisatie",
            name="besluittype",
            field=models.URLField(
                blank=True,
                help_text="URL naar het besluittype waarop de autorisatie van toepassing is.",
                max_length=1000,
                verbose_name="besluittype",
            ),
        ),
        migrations.AddField(
            model_name="autorisatie",
            name="informatieobjecttype",
            field=models.URLField(
                blank=True,
                help_text="URL naar het informatieobjecttype waarop de autorisatie van toepassing is.",
                max_length=1000,
                verbose_name="informatieobjecttype",
            ),
        ),
    ]



from django.apps import AppConfig
from django.db import models

from drf_yasg import openapi
from drf_yasg.inspectors.field import (
    basic_type_info,
    model_field_to_basic_type,
    serializer_field_to_basic_type,
)
from rest_framework import serializers

from . import fields
from .serializers import DurationField

try:
    from relativedeltafield import RelativeDeltaField
except ImportError:
    RelativeDeltaField = None

# https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.2.md#data-types
# and https://github.com/OAI/OpenAPI-Specification/issues/845 - format: duration
# is collected somewhere so there's precedent
FORMAT_DURATION = "duration"


class ZDSSchemaConfig(AppConfig):
    name = "vng_api_common"

    def ready(self):
        from . import checks  # noqa

        patch_duration_type()
        register_serializer_field()


def patch_duration_type():
    def _patch(basic_types, _field_cls, format=None):
        for index, (field_cls, basic_type) in enumerate(basic_types):
            if field_cls is _field_cls:
                basic_types[index] = (_field_cls, (openapi.TYPE_STRING, format))
                break

    _patch(model_field_to_basic_type, models.DurationField, FORMAT_DURATION)
    _patch(basic_type_info, models.DurationField, FORMAT_DURATION)
    _patch(serializer_field_to_basic_type, serializers.DurationField, FORMAT_DURATION)
    _patch(basic_type_info, serializers.DurationField, FORMAT_DURATION)

    # best-effort support for relativedeltafield
    if RelativeDeltaField is not None:
        _patch(model_field_to_basic_type, RelativeDeltaField, FORMAT_DURATION)
        _patch(basic_type_info, RelativeDeltaField, FORMAT_DURATION)


def register_serializer_field():
    mapping = serializers.ModelSerializer.serializer_field_mapping
    mapping[models.fields.DurationField] = DurationField
    mapping[fields.DaysDurationField] = DurationField

    if RelativeDeltaField is not None:
        mapping[RelativeDeltaField] = DurationField



"""
Implement polymorphism at the API level.

Add support to DRF serializers for polymorphic resources. Polymorphic resources
take a certain shape depending on the value of a field, called the
discriminator. The model itself does not need to be made polymorphic, allowing
database queries to remain performant.

Usage:

>>> from vng_api_common.polymorphism import Discriminator, PolymorphicSerializer
>>> class AutorisatieBaseSerializer(PolymorphicSerializer):
...     discriminator = Discriminator(
...         discriminator_field='type',
...         mapping={
...             'value1': (
...                 'field_for_value1',
...             ),
...             'value2': (
...                 'field_for_value2',
...             ),
...         }
...     )
...
...     class Meta:
...         model = SomeModel
...         fields = (
...             'type',
...             'common_field',
...         )
...

The serializer output will then either contain ``field_for_value2`` or
``field_for_value2``, depending on the value of the field ``type``.
"""
import logging
from collections import OrderedDict
from typing import Any, Dict, Union

from django.core.exceptions import FieldDoesNotExist

from rest_framework import serializers

__all__ = ["Discriminator", "PolymorphicSerializer"]

logger = logging.getLogger(__name__)


class Discriminator:
    def __init__(
        self,
        discriminator_field: str,
        mapping: Dict[Any, Union[tuple, serializers.ModelSerializer]],
        group_field: Union[None, str] = None,
        same_model: bool = True,
    ):
        self.discriminator_field = discriminator_field
        self.mapping = mapping
        self.group_field = group_field
        self.same_model = same_model

    def to_representation(self, instance) -> OrderedDict:
        discriminator_value = getattr(instance, self.discriminator_field)
        serializer = self.mapping.get(discriminator_value)
        if serializer is None:
            return None

        representation = serializer.to_representation(instance)
        return representation

    def to_internal_value(self, data) -> OrderedDict:
        discriminator_value = data[self.discriminator_field]
        serializer = self.mapping.get(discriminator_value)
        if serializer is None:
            return None

        internal_value = serializer.to_internal_value(data)
        # if nested serializer was generated in _sanitize_discriminator name if group_field
        # was changed in the internal_value. We need to return it
        if (
            self.group_field
            and self.group_field not in internal_value
            and len(internal_value) == 1
        ):
            key, value = internal_value.popitem()
            internal_value = OrderedDict({self.group_field: value})

        return internal_value


class PolymorphicSerializerMetaclass(serializers.SerializerMetaclass):
    @classmethod
    def _sanitize_discriminator(cls, name, attrs) -> Union[Discriminator, None]:
        discriminator = attrs["discriminator"]
        if discriminator is None:
            return None

        model = attrs["Meta"].model

        try:
            field = model._meta.get_field(discriminator.discriminator_field)
        except FieldDoesNotExist as exc:
            raise FieldDoesNotExist(
                f"The discriminator field '{discriminator.discriminator_field}' "
                f"does not exist on the model '{model._meta.label}'"
            ) from exc

        values_seen = set()

        for value, fields in discriminator.mapping.items():
            # construct a serializer instance if a tuple/list of fields is passed
            if isinstance(fields, (tuple, list)):
                name = f"{value}{model._meta.object_name}Serializer"

                Meta = type("Meta", (), {"model": model, "fields": tuple(fields)})

                serializer_class = type(
                    name, (serializers.ModelSerializer,), {"Meta": Meta}
                )

                discriminator.mapping[value] = serializer_class()

            values_seen.add(value)

            serializer = discriminator.mapping[value]

            if serializer is None:
                continue

            # rewrite it to nested serializer
            if discriminator.group_field:
                group_name = (
                    f"{discriminator.group_field}_{serializer.__class__.__name__}"
                )
                group_meta = type(
                    "Meta", (), {"model": model, "fields": (discriminator.group_field,)}
                )

                # find source field for nested serializer
                source = None
                related_fields = model._meta.fields_map
                for field_name, field_type in related_fields.items():
                    if field_type.related_model == serializer.Meta.model:
                        source = field_name

                group_field = serializer.__class__(
                    source=source, required=False, label=discriminator.group_field
                )

                group_serializer_class = type(
                    group_name,
                    (serializers.ModelSerializer,),
                    {"Meta": group_meta, discriminator.group_field: group_field},
                )
                discriminator.mapping[value] = group_serializer_class()

        if field.choices:
            values = {choice[0] for choice in field.choices}
            difference = values - values_seen
            if difference:
                logger.warn(
                    "'%s': not all possible values map to a serializer. Missing %s",
                    name,
                    difference,
                )

        return discriminator

    def __new__(cls, name, bases, attrs):
        attrs["discriminator"] = cls._sanitize_discriminator(name, attrs)
        return super().__new__(cls, name, bases, attrs)


class PolymorphicSerializer(
    serializers.HyperlinkedModelSerializer, metaclass=PolymorphicSerializerMetaclass
):
    discriminator: Discriminator = None

    def to_representation(self, instance):
        representation = super().to_representation(instance)
        extra_fields = self.discriminator.to_representation(instance)
        if extra_fields:
            representation.update(extra_fields)
        return representation

    def to_internal_value(self, data):
        internal_value = super().to_internal_value(data)
        extra_fields = self.discriminator.to_internal_value(data)
        if extra_fields:
            internal_value.update(extra_fields)
        return internal_value



# Generated by Django 2.2.2 on 2019-07-12 14:43

import uuid

import django.contrib.postgres.fields.jsonb
import django.core.serializers.json
import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("audittrails", "0008_audittrail_resource_weergave")]

    operations = [
        migrations.AlterField(
            model_name="audittrail",
            name="aanmaakdatum",
            field=models.DateTimeField(
                auto_now=True, help_text="De datum waarop de handeling is gedaan."
            ),
        ),
        migrations.AlterField(
            model_name="audittrail",
            name="actie",
            field=models.CharField(
                help_text="De uitgevoerde handeling.", max_length=50
            ),
        ),
        migrations.AlterField(
            model_name="audittrail",
            name="actie_weergave",
            field=models.CharField(
                blank=True, help_text="Vriendelijke naam van de actie.", max_length=200
            ),
        ),
        migrations.AlterField(
            model_name="audittrail",
            name="applicatie_id",
            field=models.CharField(
                blank=True,
                help_text="Unieke identificatie van de applicatie, binnen de organisatie.",
                max_length=100,
            ),
        ),
        migrations.AlterField(
            model_name="audittrail",
            name="applicatie_weergave",
            field=models.CharField(
                blank=True,
                help_text="Vriendelijke naam van de applicatie.",
                max_length=200,
            ),
        ),
        migrations.AlterField(
            model_name="audittrail",
            name="bron",
            field=models.CharField(
                choices=[
                    ("ac", "Autorisatiecomponent"),
                    ("nrc", "Notificatierouteringcomponent"),
                    ("zrc", "Zaakregistratiecomponent"),
                    ("ztc", "Zaaktypecatalogus"),
                    ("drc", "Documentregistratiecomponent"),
                    ("brc", "Besluitregistratiecomponent"),
                ],
                help_text="De naam van het component waar de wijziging in is gedaan.",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="audittrail",
            name="gebruikers_id",
            field=models.CharField(
                blank=True,
                help_text="Unieke identificatie van de gebruiker die binnen de organisatie herleid kan worden naar een persoon.",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="audittrail",
            name="gebruikers_weergave",
            field=models.CharField(
                blank=True,
                help_text="Vriendelijke naam van de gebruiker.",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="audittrail",
            name="hoofd_object",
            field=models.URLField(
                help_text="De URL naar het hoofdobject van een component.",
                max_length=1000,
            ),
        ),
        migrations.AlterField(
            model_name="audittrail",
            name="nieuw",
            field=django.contrib.postgres.fields.jsonb.JSONField(
                encoder=django.core.serializers.json.DjangoJSONEncoder,
                help_text="Volledige JSON body van het object na de actie.",
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="audittrail",
            name="oud",
            field=django.contrib.postgres.fields.jsonb.JSONField(
                encoder=django.core.serializers.json.DjangoJSONEncoder,
                help_text="Volledige JSON body van het object zoals dat bestond voordat de actie heeft plaatsgevonden.",
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="audittrail",
            name="resource",
            field=models.CharField(
                help_text="Het type resource waarop de actie gebeurde.", max_length=50
            ),
        ),
        migrations.AlterField(
            model_name="audittrail",
            name="resource_url",
            field=models.URLField(help_text="De URL naar het object.", max_length=1000),
        ),
        migrations.AlterField(
            model_name="audittrail",
            name="resource_weergave",
            field=models.CharField(
                help_text="Vriendelijke identificatie van het object.", max_length=200
            ),
        ),
        migrations.AlterField(
            model_name="audittrail",
            name="resultaat",
            field=models.IntegerField(
                help_text="HTTP status code van de API response van de uitgevoerde handeling.",
                validators=[
                    django.core.validators.MinValueValidator(100),
                    django.core.validators.MaxValueValidator(599),
                ],
            ),
        ),
        migrations.AlterField(
            model_name="audittrail",
            name="toelichting",
            field=models.TextField(
                blank=True,
                help_text="Toelichting waarom de handeling is uitgevoerd.",
                verbose_name="toelichting",
            ),
        ),
        migrations.AlterField(
            model_name="audittrail",
            name="uuid",
            field=models.UUIDField(
                default=uuid.uuid4,
                help_text="Unieke identificatie van de audit regel.",
                unique=True,
            ),
        ),
    ]



from typing import Union
from urllib.parse import urlsplit, urlunsplit

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.db.models.functions import Length
from django.utils.module_loading import import_string
from django.utils.translation import ugettext_lazy as _

from rest_framework.reverse import reverse
from solo.models import SingletonModel
from zds_client import Client, ClientAuth


class APIMixin:
    """
    Determine the absolute URL of a resource in the API.

    Model mixin that reverses the URL-path in the API based on the
    ``uuid``-field of a model instance.
    """

    def get_absolute_api_url(self, request=None, **kwargs) -> str:
        """
        Build the absolute URL of the object in the API.
        """
        # build the URL of the informatieobject
        resource_name = self._meta.model_name

        reverse_kwargs = {"uuid": self.uuid}
        reverse_kwargs.update(**kwargs)

        url = reverse(f"{resource_name}-detail", kwargs=reverse_kwargs, request=request)
        return url


class JWTSecret(models.Model):
    """
    Store credentials of clients that want to access our API.

    Only clients that are known can access the API (if so configured).
    """

    identifier = models.CharField(
        _("client ID"),
        max_length=50,
        unique=True,
        help_text=_(
            "Client ID to identify external API's and applications that access this API."
        ),
    )
    secret = models.CharField(
        _("secret"), max_length=255, help_text=_("Secret belonging to the client ID.")
    )

    class Meta:
        verbose_name = _("client credential")
        verbose_name_plural = _("client credentials")

    def __str__(self):
        return self.identifier


class APICredential(models.Model):
    """
    Store credentials for external APIs.

    When we need to authenticate against a remote API, we need to know which
    client ID and secret to use to sign the JWT.
    """

    api_root = models.URLField(
        _("API-root"),
        unique=True,
        help_text=_(
            "URL of the external API, ending in a trailing slash. Example: https://example.com/api/v1/"
        ),
    )
    label = models.CharField(
        _("label"),
        max_length=100,
        default="",
        help_text=_("Human readable label of the external API."),
    )
    client_id = models.CharField(
        _("client ID"),
        max_length=255,
        help_text=_("Client ID to identify this API at the external API."),
    )
    secret = models.CharField(
        _("secret"), max_length=255, help_text=_("Secret belonging to the client ID.")
    )
    user_id = models.CharField(
        _("user ID"),
        max_length=255,
        help_text=_(
            "User ID to use for the audit trail. Although these external API credentials are typically used by"
            "this API itself instead of a user, the user ID is required."
        ),
    )
    user_representation = models.CharField(
        _("user representation"),
        max_length=255,
        default="",
        help_text=_("Human readable representation of the user."),
    )

    class Meta:
        verbose_name = _("external API credential")
        verbose_name_plural = _("external API credentials")

    def __str__(self):
        return self.api_root

    @classmethod
    def get_auth(cls, url: str, **kwargs) -> Union[ClientAuth, None]:
        split_url = urlsplit(url)
        scheme_and_domain = urlunsplit(split_url[:2] + ("", "", ""))

        candidates = (
            cls.objects.filter(api_root__startswith=scheme_and_domain)
            .annotate(api_root_length=Length("api_root"))
            .order_by("-api_root_length")
        )

        # select the one matching
        for candidate in candidates.iterator():
            if url.startswith(candidate.api_root):
                credentials = candidate
                break
        else:
            return None

        auth = ClientAuth(
            client_id=credentials.client_id,
            secret=credentials.secret,
            user_id=credentials.user_id,
            user_representation=credentials.user_representation,
            **kwargs,
        )
        return auth


class ClientConfig(SingletonModel):
    api_root = models.URLField(_("api root"), unique=True)

    class Meta:
        abstract = True

    def __str__(self):
        return self.api_root

    def save(self, *args, **kwargs):
        if not self.api_root.endswith("/"):
            self.api_root = f"{self.api_root}/"
        super().save(*args, **kwargs)

    @classmethod
    def get_client(cls) -> Client:
        """
        Construct a client, prepared with the required auth.
        """
        config = cls.get_solo()
        Client = import_string(settings.ZDS_CLIENT_CLASS)

        api_root = config.api_root
        if not api_root:
            raise ImproperlyConfigured(
                f"Configure the API root in '{cls._meta.verbose_name}'"
            )

        if not api_root.endswith("/"):
            api_root = f"{api_root}/"

        client = Client.from_url(api_root)
        client.base_url = api_root
        client.auth = APICredential.get_auth(api_root)

        return client



from django.urls import path

from .views import ErrorDetailView, ScopesView

app_name = "vng_api_common"

urlpatterns = [
    path("fouten/<exception_class>/", ErrorDetailView.as_view(), name="error-detail"),
    path("scopes/", ScopesView.as_view(), name="scopes"),
]



# Generated by Django 2.1.1 on 2019-06-14 13:46

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("vng_api_common", "0004_auto_20190517_0903")]

    operations = [
        migrations.AlterField(
            model_name="apicredential",
            name="api_root",
            field=models.URLField(
                help_text="URL of the external API, ending in a trailing slash. Example: https://example.com/api/v1/",
                unique=True,
                verbose_name="API-root",
            ),
        ),
        migrations.AlterField(
            model_name="apicredential",
            name="client_id",
            field=models.CharField(
                help_text="Client ID to identify this API at the external API.",
                max_length=255,
                verbose_name="client ID",
            ),
        ),
        migrations.AlterField(
            model_name="apicredential",
            name="label",
            field=models.CharField(
                default="",
                help_text="Human readable label of the external API.",
                max_length=100,
                verbose_name="label",
            ),
        ),
        migrations.AlterField(
            model_name="apicredential",
            name="secret",
            field=models.CharField(
                help_text="Secret belonging to the client ID.",
                max_length=255,
                verbose_name="secret",
            ),
        ),
        migrations.AlterField(
            model_name="apicredential",
            name="user_id",
            field=models.CharField(
                help_text="User ID to use for the audit trail. Although these external API credentials are typically used bythis API itself instead of a user, the user ID is required.",
                max_length=255,
                verbose_name="user ID",
            ),
        ),
        migrations.AlterField(
            model_name="apicredential",
            name="user_representation",
            field=models.CharField(
                default="",
                help_text="Human readable representation of the user.",
                max_length=255,
                verbose_name="user representation",
            ),
        ),
        migrations.AlterField(
            model_name="jwtsecret",
            name="identifier",
            field=models.CharField(
                help_text="Client ID to identify external API's and applications that access this API.",
                max_length=50,
                unique=True,
                verbose_name="client ID",
            ),
        ),
        migrations.AlterField(
            model_name="jwtsecret",
            name="secret",
            field=models.CharField(
                help_text="Secret belonging to the client ID.",
                max_length=255,
                verbose_name="secret",
            ),
        ),
    ]



from typing import Union

from django.http import HttpRequest

try:
    from raven.contrib.django.raven_compat.models import client as sentry_client
except ImportError:

    class Client:
        def captureException(self):
            pass

    sentry_client = Client()


def get_header(request: HttpRequest, header: str) -> Union[None, str]:
    """
    Extract the value of the header from the request.

    Django 2.2 exposes request.headers, while older versions require you to
    access it from request.META with the ``HTTP_`` prefix.
    """
    # django 2.2
    if hasattr(request, "headers"):
        return request.headers.get(header)

    # older versions
    header_key = f"HTTP_{header}".replace("-", "_").upper()
    return request.META.get(header_key)



from django.db import models
from django.utils.translation import ugettext_lazy as _

from vng_api_common.descriptors import GegevensGroepType


class Group(models.Model):
    subgroup_field_1 = models.CharField(max_length=50, blank=True)
    subgroup_field_2 = models.CharField(max_length=50, blank=True, default="baz")

    subgroup = GegevensGroepType(
        {"field_1": subgroup_field_1, "field_2": subgroup_field_2},
        optional=("field_2",),
    )


class Person(models.Model):
    name = models.CharField(_("name"), max_length=50)

    address_street = models.CharField(_("street name"), max_length=255)
    address_number = models.CharField(_("house number"), max_length=10)

    address = GegevensGroepType({"street": address_street, "number": address_number})

    group = models.ForeignKey("Group", null=True, on_delete=models.CASCADE)



from django.urls import include, path

urlpatterns = [
    path("api/", include("vng_api_common.notifications.api.urls")),
    path("api/", include("notifications.api.urls")),
    path("_dummy", include("vng_api_common.urls")),
]



# Generated by Django 2.2.4 on 2019-09-18 13:35

import django.contrib.postgres.indexes
from django.contrib.postgres.operations import TrigramExtension
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("audittrails", "0010_audittrail_request_id")]

    operations = [
        TrigramExtension(),
        migrations.AddIndex(
            model_name="audittrail",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["hoofd_object"],
                name="audittrail_hoofdobject_trgm",
                opclasses=["gin_trgm_ops"],
            ),
        ),
    ]



# Generated by Django 2.0.13 on 2019-05-20 14:50

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("audittrails", "0004_auto_20190520_1238")]

    operations = [
        migrations.AlterField(
            model_name="audittrail",
            name="bron",
            field=models.CharField(
                choices=[
                    ("AC", "Autorisatiecomponent"),
                    ("NRC", "Notificatierouteringcomponent"),
                    ("ZRC", "Zaakregistratiecomponent"),
                    ("ZTC", "Zaaktypecatalogus"),
                    ("DRC", "Documentregistratiecomponent"),
                    ("BRC", "Besluitregistratiecomponent"),
                ],
                help_text="De naam van het component waar de wijziging in is gedaan",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="audittrail",
            name="resultaat",
            field=models.IntegerField(
                help_text="HTTP status code van de API response van de uitgevoerde handeling",
                validators=[
                    django.core.validators.MinValueValidator(100),
                    django.core.validators.MaxValueValidator(599),
                ],
            ),
        ),
    ]
